/*                                                                              *
 *   This file is part of the ESO IRPLIB package                                *
 *   Copyright (C) 2004,2005 European Southern Observatory                      *
 *                                                                              *
 *   This library is free software; you can redistribute it and/or modify       *
 *   it under the terms of the GNU General Public License as published by       *
 *   the Free Software Foundation; either version 2 of the License, or          *
 *   (at your option) any later version.                                        *
 *                                                                              *
 *   This program is distributed in the hope that it will be useful,            *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the              *
 *   GNU General Public License for more details.                               *
 *                                                                              *
 *   You should have received a copy of the GNU General Public License          *
 *   along with this program; if not, write to the Free Software                *
 *   Foundation, 51 Franklin St, Fifth Floor, Boston, MA  02111-1307  USA       *
 *                                                                              */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*----------------------------------------------------------------------------
                                Includes
 ----------------------------------------------------------------------------*/


#include <irplib_detmon.h>


#include <math.h>
/*---------------------------------------------------------------------------*/
/*
 * @defgroup irplib_detmon_test   Testing of the IRPLIB utilities
 */
/*---------------------------------------------------------------------------*/

#define PI 3.141592653
/*----------------------------------------------------------------------------
                                Defines
 ----------------------------------------------------------------------------*/

#ifndef IMAGESIZE
#define IMAGESIZE       128
#endif

#define IMAGELISTSIZE   10

#define NBADPIXS        10

#undef REGEXP
#define REGEXP                            "^(ARCFILE|MJD-OBS|ESO TPL ID|" \
                                          "DATE-OBS|ESO DET DIT|ESO DET NDIT|"\
                                          "ESO DET NCORRS|"\
                                          "ESO DET MODE NAME)$"

/*----------------------------------------------------------------------------
                            Private Function prototypes
 ----------------------------------------------------------------------------*/

static void irplib_detmon_tests(void);


cpl_frameset * irplib_detmon_pernoise_tests_fill_frameset(cpl_boolean,
                                                          int);

cpl_frameset * irplib_detmon_tests_fill_frameset(const char     * tag_on,
                                                 const char     * tag_off,
                                                 int              non,
                                                 cpl_boolean      opt_nir,
                                                 cpl_boolean      with_exts,
                                                 int              nexts);
static cpl_error_code
irplib_ksigma_clip_bf(cpl_image * img,
		      const int llx,
		      const int lly,
		      const int urx, int ury,
		      const int kappa,
		      const int nclip,
		      double *clean_mean, double *clean_stdev);

/*---------------------------------------------------------------------------*/
/*
 * @brief   Unit tests of fit module
 */
/*---------------------------------------------------------------------------*/

int main(void)
{
    /* Initialize CPL + IRPLIB */
    cpl_test_init(PACKAGE_BUGREPORT, CPL_MSG_WARNING);

    irplib_detmon_tests();

    return cpl_test_end(0);
}

static void irplib_detmon_tests(void)
{

    cpl_parameterlist   * parlist;
    cpl_frameset        * frameset;
    cpl_frameset        * set;

    const char * tag = "RAW";
    cpl_error_code error;

    cpl_propertylist * prolist;
    /* HIGH LEVEL FUNCTION FOR RON/BIAS RECIPE */

    frameset = irplib_detmon_tests_fill_frameset(tag, tag, 3, TRUE, FALSE, 0);
    cpl_test(frameset != NULL);
    set = cpl_frameset_duplicate(frameset);
    cpl_test(set != NULL);
    parlist = cpl_parameterlist_new();
    cpl_test(parlist != NULL);
    cpl_test_error(CPL_ERROR_NONE);
    prolist = irplib_detmon_fill_prolist("PROCATG", "TYPE", "TECH", CPL_TRUE);

    irplib_detmon_ronbias_fill_parlist_default(parlist, "detmon_ronbias", "detmon");


    irplib_detmon_ronbias(set, parlist, tag, "detmon_ronbias", "detmon",
			  REGEXP, prolist, NULL, NULL, prolist, prolist,
			  prolist, prolist,  "DETMON / 1.0", NULL, OPT);

    cpl_test_error(CPL_ERROR_NONE);
    cpl_error_reset();

    cpl_parameterlist_delete(parlist);
    cpl_frameset_delete(set);

    /* 2. test with extensions */
    set = irplib_detmon_tests_fill_frameset(tag, tag, 3, TRUE, TRUE, 4);
    parlist = cpl_parameterlist_new();
	irplib_detmon_ronbias_fill_parlist(parlist, "detmon_ronbias", "detmon",
					   "ALL", /* --method */
					   "NORM",/* --pmethod */
					   1,     /* --preoverscan_degree */
					   -1,    /* --random_nsamples */
					   -1,    /* --random_sizex */
					   -1,    /* --random_sizey */
					   0,     /* --criteria     */
					   -1,    /* --ref_llx     */
					   -1,    /* --ref_lly */
					   -1,    /* --ref_urx */
					   -1,    /* --ref_ury */
					   "MEAN",/* --stacking_method */
					   3,     /* --stacking_ks_low */
					   10000, /* --stacking_ks_high */
					   25,    /* --stacking_ks_iter */
					   0,     /* --master_shift_x */
					   0,     /* --master_shift_y */
					   -1,    /* --ron_llx */
					   -1,    /* --ron_lly */
					   -1,    /* --ron_urx */
					   -1,    /* --ron_ury */
					   -1,    /* --exts */
					   OPT);

    error =irplib_detmon_ronbias(set, parlist, tag, "detmon_ronbias", "detmon",
				 REGEXP, prolist, NULL, NULL, prolist, prolist,
				 prolist, prolist,  "DETMON / 1.0", NULL, OPT);

    /* Move the NULLs away when X and Y structure built */

    cpl_test_eq(error, CPL_ERROR_NONE);
    cpl_test_error(CPL_ERROR_NONE);

    cpl_parameterlist_delete(parlist);
    cpl_frameset_delete(set);

    cpl_frameset_delete(frameset);
    cpl_propertylist_delete(prolist);


#ifdef HAVE_FFTW
    /* Tests for Periodic Noise Characterisation */

    frameset = irplib_detmon_pernoise_tests_fill_frameset(0,0);

    parlist = cpl_parameterlist_new();
    irplib_detmon_fill_pernoise_params_default(parlist,
                                               "detmon_pernoise", "detmon");

    irplib_detmon_pernoise(frameset, parlist, "RAW", "detmon_pernoise", "detmon", "PRO-1.15", "DETMON / 1.0", NULL);

    cpl_parameterlist_delete(parlist);
    cpl_frameset_delete(frameset);

#endif

    /* Test for DARK recipe */

    frameset = irplib_detmon_tests_fill_frameset("DARK", "DARK", 20,FALSE, FALSE, 0);
    parlist = cpl_parameterlist_new();
    error = irplib_detmon_fill_dark_params_default(parlist,
                                               "detmon_dark", "detmon");
    cpl_test_zero(error);

    irplib_detmon_dark(frameset, parlist, "DARK", "detmon_dark", "detmon",
                       "MASTER", "DSNU", "DSNU_TABLE", "DETMON / 1.0", NULL);

    cpl_parameterlist_delete(parlist);
    cpl_frameset_delete(frameset);

    /* New test for irplib_imagelist_collapse_stdev_create() */
    {
	cpl_imagelist * mylist = cpl_imagelist_new();
	cpl_image     * image  = cpl_image_new(IMAGESIZE, IMAGESIZE,
					       CPL_TYPE_FLOAT);
	cpl_image     * collapsed;
	double          mean, stdev;

	int i;
	for (i = 1; i <= 5; i++) {
	    cpl_image * to_list =
		cpl_image_add_scalar_create(image, 10.0 * i);
	    cpl_imagelist_set(mylist, to_list, i - 1);
	}

	collapsed = irplib_imagelist_collapse_stdev_create(mylist);
	mean  = cpl_image_get_mean(collapsed);
	stdev = cpl_image_get_stdev(collapsed);

	cpl_test_rel(mean, 15.81138, 1e-4);
	cpl_test_zero(stdev);

	cpl_imagelist_delete(mylist);
	cpl_image_delete(image);
	cpl_image_delete(collapsed);
    }

    /* Tests for ksigma_clip() */
    {
	cpl_image     * image  = cpl_image_new(IMAGESIZE, IMAGESIZE,
					       CPL_TYPE_FLOAT);
	double mean=0;
        double stdev=0;
	double mean_bf=0;
        double stdev_bf=0;

	cpl_image_fill_noise_uniform(image, 300, 350);

	cpl_image_set(image, 50, 50, 0.0);
	cpl_image_set(image, 15, 1, 0.0);
	cpl_image_set(image, 76, 32, 0.0);
	cpl_image_set(image, 10, 19, 0.0);

	cpl_image_set(image, 30, 10, 6000.0);
	cpl_image_set(image, 31, 11, 6020.0);
	cpl_image_set(image, 30, 20, 6030.0);
	cpl_image_set(image, 30, 40, 6040.0);

	error = irplib_ksigma_clip(image, 1, 1, IMAGESIZE, IMAGESIZE,
				   3.0, 10, 1e-5, &mean, &stdev);
	cpl_test(error == CPL_ERROR_NONE);
	cpl_test(mean  > 320.0 && mean  < 330.0);
	cpl_test(stdev > 12.0  && stdev < 18.0 );


	error = irplib_ksigma_clip_bf(image, 1, 1, IMAGESIZE, IMAGESIZE,
				      3.0, 10, &mean_bf, &stdev_bf);

	cpl_test(abs(mean  - mean_bf)  < FLT_EPSILON);
	cpl_test(abs(stdev - stdev_bf) < FLT_EPSILON);

	cpl_image_delete(image);
    }

    cpl_test_zero(system("rm *.paf *.fits"));
}


cpl_frameset * irplib_detmon_pernoise_tests_fill_frameset(cpl_boolean with_exts,
                                                          int nexts)
{

    cpl_frameset                * frameset = cpl_frameset_new();
    cpl_frame                   * frame;
    cpl_propertylist            * plist;

    /* Create dummy image(s) */

    cpl_image * raw1 = cpl_image_new(512, 512, CPL_TYPE_FLOAT);
    cpl_image * raw2 = cpl_image_new(512, 512, CPL_TYPE_FLOAT);
    int i, j, k = 0;

    cpl_image_fill_noise_uniform(raw1, -2, 2);

    cpl_image_add_scalar(raw1, 100);


    for(i = 1; i <= 512; i++){
        for(j = 1; j <= 512; j++){
            int rejected;
            double value = cpl_image_get(raw1, j, i, &rejected);
            cpl_image_set(raw1, j, i, 4*sin((double)i/16*CPL_MATH_2PI) + value);
            if(rejected < 0) {
               cpl_msg_error(cpl_func, "Error adding test noise");
               k++;
            }
        }
    }

    cpl_image_fill_noise_uniform(raw2, -3, 3);

    frame = cpl_frame_new();
    cpl_frame_set_filename(frame, "pertest1.fits");
    cpl_frame_set_tag(frame, "RAW");
    cpl_frame_set_type(frame, CPL_FRAME_TYPE_IMAGE);
    cpl_frame_set_group(frame, CPL_FRAME_GROUP_RAW);

    cpl_frameset_insert(frameset, frame);

    plist = cpl_propertylist_new();

    cpl_propertylist_append_double(plist, "ESO DET DIT", (double)(3));
    cpl_propertylist_append_double(plist, "EXPTIME", (double)(3));
    cpl_propertylist_append_double(plist, "MJD-OBS", 0);


    if(with_exts == TRUE) {
        cpl_image_save(NULL, "pertest1.fits", CPL_BPP_IEEE_FLOAT, plist, CPL_IO_DEFAULT);
        for(j = 0; j < nexts; j++) {
            cpl_image_save(raw1, "pertest1.fits", CPL_BPP_IEEE_FLOAT, plist, CPL_IO_EXTEND);
        }
    } else {
        cpl_image_save(raw1, "pertest1.fits", CPL_BPP_IEEE_FLOAT, plist, CPL_IO_DEFAULT);
    }


    frame = cpl_frame_new();
    cpl_frame_set_filename(frame, "pertest2.fits");
    cpl_frame_set_tag(frame, "RAW");
    cpl_frame_set_type(frame, CPL_FRAME_TYPE_IMAGE);
    cpl_frame_set_group(frame, CPL_FRAME_GROUP_RAW);

    cpl_frameset_insert(frameset, frame);

    if(with_exts == TRUE) {
        cpl_image_save(NULL, "pertest2.fits", CPL_BPP_IEEE_FLOAT, plist, CPL_IO_DEFAULT);
        for(j = 0; j < nexts; j++) {
            cpl_image_save(raw2, "pertest2.fits", CPL_BPP_IEEE_FLOAT, plist, CPL_IO_EXTEND);
        }
    } else {
        cpl_image_save(raw2, "pertest2.fits", CPL_BPP_IEEE_FLOAT, plist, CPL_IO_DEFAULT);
    }
    cpl_propertylist_delete(plist);
    cpl_image_delete(raw2);
    cpl_image_delete(raw1);

    return frameset;
}

/*---------------------------------------------------------------------------*/
/*
 * @brief  Detect bad pixels of a selected type
 * @param  in           Input image
 * @param  mean         Mean level of the image - used to compute threshold
 * @param  sigma        Stdev level of the image - used to compute threshold
 * @param  flag         Flag - represents selected type of bad pixels (see enum)
 * @return The detected bad pixels in a cpl_mask.
 */
/*---------------------------------------------------------------------------*/

cpl_frameset * irplib_detmon_tests_fill_frameset(const char     * tag_on,
                                                 const char     * tag_off,
                                                 int              non,
                                                 cpl_boolean      opt_nir,
                                                 cpl_boolean      with_exts,
                                                 int              nexts)
{
    cpl_frameset                * frameset;
    cpl_frame                   * frame;
    int i;
    char                        frame_name[20];
    cpl_propertylist            * plist;
    cpl_image                   * image;
    frameset = cpl_frameset_new();

    for (i=0; i<non; i++) {
        frame = cpl_frame_new();
        sprintf(frame_name, "dummyon%d.fits",i);
        cpl_frame_set_filename(frame, frame_name);
        cpl_frame_set_tag(frame, tag_on);
        cpl_frame_set_type(frame, CPL_FRAME_TYPE_IMAGE);
        cpl_frame_set_group(frame, CPL_FRAME_GROUP_RAW);

        cpl_frameset_insert(frameset, frame);

        plist = cpl_propertylist_new();
        if(opt_nir == TRUE) {
            cpl_propertylist_append_double(plist, "ESO DET DIT", (double)(i/2+1));
            cpl_propertylist_append_double(plist, "EXPTIME", (double)(i/2+1));
            cpl_propertylist_append_double(plist, "MJD-OBS", 100);

	    cpl_propertylist_append_int(plist, "NAXIS1", IMAGESIZE);
	    cpl_propertylist_append_int(plist, "NAXIS2", IMAGESIZE);
	    cpl_propertylist_append_int(plist, "ESO DET OUT1 NX", IMAGESIZE + 10 );
	    cpl_propertylist_append_int(plist, "ESO DET OUT1 NY", IMAGESIZE);
	    cpl_propertylist_append_int(plist, "ESO DET OUT1 PRSCX", 5);
	    cpl_propertylist_append_int(plist, "ESO DET OUT1 OVSCX", 5);
        } else {
            cpl_propertylist_append_double(plist, "ESO DET DIT", (double)((i+1)/2));
            cpl_propertylist_append_double(plist, "EXPTIME", (double)((i+1)/2));
            cpl_propertylist_append_double(plist, "MJD-OBS",100);

	    cpl_propertylist_append_int(plist, "NAXIS1", IMAGESIZE);
	    cpl_propertylist_append_int(plist, "NAXIS2", IMAGESIZE);
	    cpl_propertylist_append_int(plist, "ESO DET OUT1 NX", IMAGESIZE + 10 );
	    cpl_propertylist_append_int(plist, "ESO DET OUT1 NY", IMAGESIZE);
	    cpl_propertylist_append_int(plist, "ESO DET OUT1 PRSCX", 5);
	    cpl_propertylist_append_int(plist, "ESO DET OUT1 OVSCX", 5);

        }

        if(with_exts == TRUE) {
          int j;
          cpl_image_save(NULL, frame_name, CPL_BPP_IEEE_FLOAT, plist, CPL_IO_DEFAULT);
          for(j = 0; j < nexts; j++) {
            image = cpl_image_new(IMAGESIZE, IMAGESIZE, CPL_TYPE_FLOAT);
            cpl_image_fill_noise_uniform(image, 25, 50);
            cpl_ensure(cpl_error_get_code()== CPL_ERROR_NONE, CPL_ERROR_NULL_INPUT, NULL);
	    cpl_image_add_scalar(image, 300);

	    cpl_image_set(image, 10, 10, 0.0);
	    cpl_image_set(image, 11, 11, 0.0);
	    cpl_image_set(image, 20, 20, 0.0);
	    cpl_image_set(image, 10, 40, 0.0);
	    cpl_image_set(image, 40, 10, 0.0);
	    cpl_image_set(image, 12, 32, 0.0);
	    cpl_image_set(image, 50, 50, 0.0);
	    cpl_image_set(image, 15, 1, 0.0);
	    cpl_image_set(image, 76, 32, 0.0);
	    cpl_image_set(image, 10, 19, 0.0);

	    cpl_image_set(image, 30, 10, 6000.0);
	    cpl_image_set(image, 31, 11, 6020.0);
	    cpl_image_set(image, 30, 20, 6030.0);
	    cpl_image_set(image, 30, 40, 6040.0);
	    cpl_image_set(image, 30, 10, 6070.0);
	    cpl_image_set(image, 52, 32, 6300.0);
	    cpl_image_set(image, 60, 50, 6500.0);
	    cpl_image_set(image, 85, 1, 6100.0);
	    cpl_image_set(image, 86, 32, 6040.0);
	    cpl_image_set(image, 90, 19, 6020.0);

            cpl_image_save(image, frame_name, CPL_BPP_IEEE_FLOAT, plist, CPL_IO_EXTEND);
            cpl_image_delete(image);
          }
        } else {
            image = cpl_image_new(IMAGESIZE, IMAGESIZE, CPL_TYPE_FLOAT);
            cpl_image_fill_noise_uniform(image, 25, 50);
            cpl_ensure(cpl_error_get_code()== CPL_ERROR_NONE, CPL_ERROR_NULL_INPUT, NULL);
	    cpl_image_add_scalar(image, 300);
	    cpl_image_set(image, 10, 10, 0.0);
	    cpl_image_set(image, 11, 11, 0.0);
	    cpl_image_set(image, 20, 20, 0.0);
	    cpl_image_set(image, 10, 40, 0.0);
	    cpl_image_set(image, 40, 10, 0.0);
	    cpl_image_set(image, 12, 32, 0.0);
	    cpl_image_set(image, 50, 50, 0.0);
	    cpl_image_set(image, 15, 1, 0.0);
	    cpl_image_set(image, 76, 32, 0.0);
	    cpl_image_set(image, 10, 19, 0.0);

	    cpl_image_set(image, 30, 10, 6000.0);
	    cpl_image_set(image, 31, 11, 6020.0);
	    cpl_image_set(image, 30, 20, 6030.0);
	    cpl_image_set(image, 30, 40, 6040.0);
	    cpl_image_set(image, 30, 10, 6070.0);
	    cpl_image_set(image, 52, 32, 6300.0);
	    cpl_image_set(image, 60, 50, 6500.0);
	    cpl_image_set(image, 85, 1, 6100.0);
	    cpl_image_set(image, 86, 32, 6040.0);
	    cpl_image_set(image, 90, 19, 6020.0);

            cpl_image_save(image, frame_name, CPL_BPP_IEEE_FLOAT, plist, CPL_IO_DEFAULT);


            cpl_image_delete(image);
        }
        cpl_propertylist_delete(plist);


        frame = cpl_frame_new();
        sprintf(frame_name, "dummyoff%d.fits",i);
        cpl_frame_set_filename(frame, frame_name);
        cpl_frame_set_tag(frame, tag_off);
        cpl_frame_set_type(frame, CPL_FRAME_TYPE_IMAGE);
        cpl_frame_set_group(frame, CPL_FRAME_GROUP_RAW);

        cpl_frameset_insert(frameset, frame);

        plist = cpl_propertylist_new();
        if(opt_nir == TRUE) {
            cpl_propertylist_append_double(plist, "ESO DET DIT", (double)(i/2+1));
            cpl_propertylist_append_double(plist, "EXPTIME", (double)(i/2+1));
            cpl_propertylist_append_double(plist, "MJD-OBS", 100);

	    cpl_propertylist_append_int(plist, "NAXIS1", IMAGESIZE);
	    cpl_propertylist_append_int(plist, "NAXIS2", IMAGESIZE);
	    cpl_propertylist_append_int(plist, "ESO DET OUT1 NX", IMAGESIZE + 10 );
	    cpl_propertylist_append_int(plist, "ESO DET OUT1 NY", IMAGESIZE);
	    cpl_propertylist_append_int(plist, "ESO DET OUT1 PRSCX", 5);
	    cpl_propertylist_append_int(plist, "ESO DET OUT1 OVSCX", 5);
        } else {
            cpl_propertylist_append_double(plist, "ESO DET DIT", (double)((i+1)/2));
            cpl_propertylist_append_double(plist, "EXPTIME", (double)((i+1)/2));
            cpl_propertylist_append_double(plist, "MJD-OBS", 100);

	    cpl_propertylist_append_int(plist, "NAXIS1", IMAGESIZE);
	    cpl_propertylist_append_int(plist, "NAXIS2", IMAGESIZE);
	    cpl_propertylist_append_int(plist, "ESO DET OUT1 NX", IMAGESIZE + 10 );
	    cpl_propertylist_append_int(plist, "ESO DET OUT1 NY", IMAGESIZE);
	    cpl_propertylist_append_int(plist, "ESO DET OUT1 PRSCX", 5);
	    cpl_propertylist_append_int(plist, "ESO DET OUT1 OVSCX", 5);
        }

        if(with_exts == TRUE) {
          int j;
          cpl_image_save(NULL, frame_name, CPL_BPP_IEEE_FLOAT, plist, CPL_IO_DEFAULT);
          for(j = 0; j < nexts; j++) {
            image = cpl_image_new(IMAGESIZE, IMAGESIZE, CPL_TYPE_FLOAT);
            cpl_image_fill_noise_uniform(image, 25, 50);
            cpl_ensure(cpl_error_get_code()== CPL_ERROR_NONE, CPL_ERROR_NULL_INPUT, NULL);
	    cpl_image_add_scalar(image, 300);
	    cpl_image_set(image, 10, 10, 0.0);
	    cpl_image_set(image, 11, 11, 0.0);
	    cpl_image_set(image, 20, 20, 0.0);
	    cpl_image_set(image, 10, 40, 0.0);
	    cpl_image_set(image, 40, 10, 0.0);
	    cpl_image_set(image, 12, 32, 0.0);
	    cpl_image_set(image, 50, 50, 0.0);
	    cpl_image_set(image, 15, 1, 0.0);
	    cpl_image_set(image, 76, 32, 0.0);
	    cpl_image_set(image, 10, 19, 0.0);

	    cpl_image_set(image, 30, 10, 6000.0);
	    cpl_image_set(image, 31, 11, 6020.0);
	    cpl_image_set(image, 30, 20, 6030.0);
	    cpl_image_set(image, 30, 40, 6040.0);
	    cpl_image_set(image, 30, 10, 6070.0);
	    cpl_image_set(image, 52, 32, 6300.0);
	    cpl_image_set(image, 60, 50, 6500.0);
	    cpl_image_set(image, 85, 1, 6100.0);
	    cpl_image_set(image, 86, 32, 6040.0);
	    cpl_image_set(image, 90, 19, 6020.0);

            cpl_image_save(image, frame_name, CPL_BPP_IEEE_FLOAT, plist, CPL_IO_EXTEND);
            cpl_image_delete(image);
          }
        } else {
            image = cpl_image_new(IMAGESIZE, IMAGESIZE, CPL_TYPE_FLOAT);
            cpl_image_fill_noise_uniform(image, 25, 50);
            cpl_ensure(cpl_error_get_code()== CPL_ERROR_NONE, CPL_ERROR_NULL_INPUT, NULL);
	    cpl_image_add_scalar(image, 300);
	    cpl_image_set(image, 10, 10, 0.0);
	    cpl_image_set(image, 11, 11, 0.0);
	    cpl_image_set(image, 20, 20, 0.0);
	    cpl_image_set(image, 10, 40, 0.0);
	    cpl_image_set(image, 40, 10, 0.0);
	    cpl_image_set(image, 12, 32, 0.0);
	    cpl_image_set(image, 50, 50, 0.0);
	    cpl_image_set(image, 15, 1, 0.0);
	    cpl_image_set(image, 76, 32, 0.0);
	    cpl_image_set(image, 10, 19, 0.0);

	    cpl_image_set(image, 30, 10, 6000.0);
	    cpl_image_set(image, 31, 11, 6020.0);
	    cpl_image_set(image, 30, 20, 6030.0);
	    cpl_image_set(image, 30, 40, 6040.0);
	    cpl_image_set(image, 30, 10, 6070.0);
	    cpl_image_set(image, 52, 32, 6300.0);
	    cpl_image_set(image, 60, 50, 6500.0);
	    cpl_image_set(image, 85, 1, 6100.0);
	    cpl_image_set(image, 86, 32, 6040.0);
	    cpl_image_set(image, 90, 19, 6020.0);

            cpl_image_save(image, frame_name, CPL_BPP_IEEE_FLOAT, plist, CPL_IO_DEFAULT);


            cpl_image_delete(image);
        }
        cpl_propertylist_delete(plist);
    }

    return frameset;
}

/**
  @brief    Get clean mean and stdev of an image over a window
  @param    img input image
  @param    llx input lower left x image's window coordinate
  @param    lly input lower left y image's window coordinate
  @param    urx input upper right y image's window coordinate
  @param    ury input upper right y image's window coordinate
  @param    kappa  input kappa of kappa-sigma clip
  @param    nclip input max no of kappa-sigma clip iterations
  @param    clean_mean output upper right y image's window coordinate
  @param    clean_stdev output upper right y image's window coordinate
  @return   pixel scale
 */

static cpl_error_code
irplib_ksigma_clip_bf(cpl_image * img,
		      const int llx,
		      const int lly,
		      const int urx, int ury,
		      const int kappa,
		      const int nclip,
		      double *clean_mean, double *clean_stdev)
{


    double                  mean = 0;
    double                  stdev = 0;
    double                  threshold = 0;
    double                  lo_cut = 0;
    double                  hi_cut = 0;
    double                  lo_cut_p = 0;
    double                  hi_cut_p = 0;

    cpl_mask               *mask = NULL;
    cpl_image              *tmp = NULL;
    cpl_stats              *stats = NULL;
    int                     i = 0;

    if((tmp = cpl_image_extract(img, llx, lly, urx, ury)) == NULL) {
        cpl_msg_error(cpl_func, "Error extracting window");
        return -1;
    }
    cpl_image_accept_all(tmp);
    for(i = 0; i < nclip; i++) {


        cpl_stats_delete(stats);
        stats =
            cpl_stats_new_from_image(tmp, CPL_STATS_MEAN | CPL_STATS_STDEV);
        mean = cpl_stats_get_mean(stats);
        stdev = cpl_stats_get_stdev(stats);

        threshold = kappa * stdev;
        lo_cut = mean - threshold;
        hi_cut = mean + threshold;

        cpl_image_accept_all(tmp);
        mask = cpl_mask_threshold_image_create(tmp, lo_cut, hi_cut);

        cpl_mask_not(mask);
        cpl_image_reject_from_mask(tmp, mask);
        cpl_mask_delete(mask);
        if(lo_cut_p == lo_cut && hi_cut_p == hi_cut)
            break;
        else {
            lo_cut_p = lo_cut;
            hi_cut_p = hi_cut;
        }

    }
    *clean_mean = mean;
    *clean_stdev = stdev;
    cpl_image_delete(tmp);
    cpl_stats_delete(stats);

    return cpl_error_get_code();


}

